/*
 *  AssociationRulesTask.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import java.io.File;
import java.io.FileReader;

import weka.associations.Associator;
import weka.core.Instances;
import weka.core.Utils;
import weka4ws.stubs.service.ModelResult;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class AssociationRulesTask extends DataMiningTask {

    public AssociationRulesTask(ServiceParameters parameters,
            Weka4WSResource resultResource) throws Exception {
        super(parameters, "associationRules", resultResource);
    }

    /**
     * Starts running the currently configured associator with the current
     * settings. This is run in a separate thread, and will only start if
     * there is no associator already running. The associator output is sent
     * to the results history panel.
     */
    protected void startAssociator() {

        if (m_RunThread == null) {
            /* =============== Weka4WS code removed ============== */
            m_RunThread = new Thread() {
                public void run() {

                    /* =============== Weka4WS code removed ============== */

                    /* ============== Weka4WS modifications ============== */
                    
                    try {
                        File datasetFile = new File(tempDataset.getFilePath());
                        FileReader datasetReader = new FileReader(datasetFile);
                        Instances inst = new Instances(datasetReader);

                        ModelResult modelResult = new ModelResult();

                        String[] optionsArray = Utils.splitOptions(algorithm
                                .getParameters());

                        Associator associator = Associator.forName(algorithm
                                .getName(), optionsArray);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        String cname = associator.getClass().getName();

                        /* =============== Weka4WS code removed ============== */

                        // Output some header information
                        /* ============== Weka4WS modifications ============== */
                        logger.info("Started " + cname);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        // Build the model and output it.
                        /* ============== Weka4WS modifications ============== */
                        logger.info("Building model on training data...");
                        /* =================================================== */

                        associator.buildAssociations(inst);

                        /* ============== Weka4WS modifications ============== */
                        modelResult.setModel(WUtil.getStrFromObj(associator));
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        logger.info("Finished " + cname);

                        result.setModel(modelResult);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                    } catch (Exception ex) {
                        /* ============== Weka4WS modifications ============== */
                        ex.printStackTrace();
                        wekaException.setThrown(true);
                        wekaException.setMessage(ex.getMessage());
                        /* =================================================== */

                        logger.debug("ex.getMessage()=" + ex.getMessage());

                    } finally {

                        /* =============== Weka4WS code removed ============== */

                        m_RunThread = null;

                        /* ============== Weka4WS modifications ============== */
                        finalOperations(result, wekaException);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                    }
                }
            };
            /* =============== Weka4WS code removed ============== */
            
            m_RunThread.start();
        }
    }

}
