/*
 *  ClassificationTask.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import java.io.File;
import java.io.FileReader;
import java.util.Random;

import weka.classifiers.Classifier;
import weka.classifiers.CostMatrix;
import weka.classifiers.Evaluation;
import weka.classifiers.evaluation.NominalPrediction;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;
import weka.gui.visualize.Plot2D;
import weka4ws.stubs.service.ArrayOfStrings;
import weka4ws.stubs.service.EvalOptions;
import weka4ws.stubs.service.EvalResult;
import weka4ws.stubs.service.ModelResult;
import weka4ws.stubs.service.TestOptions;
import weka4ws.stubs.service.VisResult;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class ClassificationTask extends DataMiningTestTask {

    public ClassificationTask(ClassificationParameters parameters,
            Weka4WSResource resultResource) throws Exception {

        super(parameters, "classification", resultResource);

    }

    /**
     * This is the same method of the class
     * <code>weka.gui.explorer.ClassifierPanel</code> taken from the source
     * code of Weka version 3.4.11
     */
    public void startClassifier() {

        if (m_RunThread == null) {

            /* =============== Weka4WS code removed ============== */

            m_RunThread = new Thread() {
                public void run() {

                    /* =============== Weka4WS code removed ============== */

                    /* ============== Weka4WS modifications ============== */

                    try {

                        ClassificationParameters classParameters = (ClassificationParameters) parameters;
                        EvalOptions evalOptions = classParameters
                                .getEvalOptions();
                        TestOptions testOptions = classParameters
                                .getTestOptions();

                        File datasetFile = new File(tempDataset.getFilePath());
                        FileReader datasetReader = new FileReader(datasetFile);
                        Instances inst = new Instances(datasetReader);

                        ModelResult modelResult = new ModelResult();
                        EvalResult evalResult = new EvalResult();
                        VisResult visResult = new VisResult();

                        /* =================================================== */

                        Instances userTest = null;
                        // additional vis info (either shape type or point size)
                        FastVector plotShape = new FastVector();
                        FastVector plotSize = new FastVector();
                        Instances predInstances = null;

                        // will hold the prediction objects if the class is
                        // nominal
                        FastVector predictions = null;

                        // for timing
                        long trainTimeStart = 0, trainTimeElapsed = 0;

                        /* ============== Weka4WS modifications ============== */
                        int numFolds = testOptions.getNumFolds();
                        int percent = testOptions.getPercent();

                        if (testMode == TEST_SET) {
                            File testsetFile = new File(tempTestset
                                    .getFilePath());
                            FileReader testsetReader = new FileReader(
                                    testsetFile);
                            userTest = new Instances(testsetReader);
                        }

                        if (evalOptions == null) {
                            throw new Exception("evalOptions is null");
                        }

                        CostMatrix costMatrix = (CostMatrix) WUtil
                                .getObjFromStr(evalOptions.getCostMatrix());
                        boolean outputModel = evalOptions.isOutputModel();
                        boolean outputConfusion = evalOptions
                                .isOutputConfusion();
                        boolean outputPerClass = evalOptions.isOutputPerClass();
                        boolean outputSummary = evalOptions.isOutputSummary();
                        boolean outputEntropy = evalOptions.isOutputEntropy();
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        int classIndex = classParameters.getClassIndex();
                        String[] optionsArray = Utils.splitOptions(algorithm
                                .getParameters());

                        Classifier classifier = Classifier.forName(algorithm
                                .getName(), optionsArray);
                        /* =================================================== */

                        Classifier template = null;
                        try {
                            template = Classifier.makeCopy(classifier);
                        } catch (Exception ex) {

                            /*
                             * ============== Weka4WS modifications
                             * ==============
                             */
                            logger.info("Problem copying classifier: "
                                    + ex.getMessage());
                            /* =================================================== */

                        }

                        /* =============== Weka4WS code removed ============== */

                        String cname = classifier.getClass().getName();

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        if (testMode == TEST_SET) {
                            userTest.setClassIndex(classIndex);
                        }
                        /* =================================================== */

                        inst.setClassIndex(classIndex);

                        // set up the structure of the plottable instances for
                        // visualization
                        predInstances = setUpVisualizableInstances(inst);
                        predInstances.setClassIndex(inst.classIndex() + 1);

                        if (inst.classAttribute().isNominal()) {
                            predictions = new FastVector();
                        }

                        // Output some header information
                        /* ============== Weka4WS modifications ============== */
                        logger.info("Started " + cname);

                        // Build the model and output it.
                        if (outputModel || (testMode == 3) || (testMode == 4)) {
                            logger.info("Building model on training data...");
                            trainTimeStart = System.currentTimeMillis();
                            classifier.buildClassifier(inst);
                            modelResult.setModel(WUtil
                                    .getStrFromObj(classifier));
                            trainTimeElapsed = System.currentTimeMillis()
                                    - trainTimeStart;
                            logger.info("Time taken to build model: "
                                    + Utils.doubleToString(
                                            trainTimeElapsed / 1000.0, 2)
                                    + " seconds");
                        }

                        int rnd = 1;
                        ArrayOfStrings classifiers;
                        /* =================================================== */

                        Evaluation eval = null;
                        switch (testMode) {
                        case 3: // Test on training
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Evaluating on training data...");
                            /* =================================================== */
                            eval = new Evaluation(inst, costMatrix);

                            /* =============== Weka4WS code removed ============== */

                            for (int jj = 0; jj < inst.numInstances(); jj++) {
                                processClassifierPrediction(inst.instance(jj),
                                        classifier, eval, predictions,
                                        predInstances, plotShape, plotSize);

                                /* =============== Weka4WS code removed ============== */

                                if ((jj % 1000) == 0) {
                                    /* ============== Weka4WS modifications ============== */
                                    logger
                                            .info("Evaluating on training data. Processed "
                                                    + jj + " instances...");
                                    /* =================================================== */
                                }
                            }

                            /* =============== Weka4WS code removed ============== */

                            break;

                        case 1: // CV mode
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Randomizing instances...");
                            rnd = evalOptions.getRnd();
                            /* =================================================== */

                            /* =============== Weka4WS code removed ============== */

                            Random random = new Random(rnd);
                            inst.randomize(random);
                            if (inst.attribute(classIndex).isNominal()) {
                                /* ============== Weka4WS modifications ============== */
                                logger.info("Stratifying instances...");
                                /* =================================================== */
                                inst.stratify(numFolds);
                            }
                            eval = new Evaluation(inst, costMatrix);

                            /* =============== Weka4WS code removed ============== */

                            /* ============== Weka4WS modifications ============== */
                            logger.info("Building and evaluating models for "
                                    + numFolds + " folds...");
                            classifiers = new ArrayOfStrings(
                                    new String[numFolds]);
                            /* =================================================== */

                            // Make some splits and do a CV
                            for (int fold = 0; fold < numFolds; fold++) {

                                /* =============== Weka4WS code removed ============== */

                                Instances train = inst.trainCV(numFolds, fold,
                                        random);
                                eval.setPriors(train);

                                /* =============== Weka4WS code removed ============== */

                                Classifier current = null;
                                try {
                                    current = Classifier.makeCopy(template);
                                } catch (Exception ex) {
                                    /* ============== Weka4WS modifications ============== */
                                    logger.error("Problem copying classifier: "
                                            + ex.getMessage());
                                    /* =================================================== */
                                }
                                logger.debug("Building classifier " + fold);
                                current.buildClassifier(train);
                                /* ============== Weka4WS modifications ============== */
                                logger.debug("Adding classifier " + fold
                                        + " to ArrayOfStrings");
                                classifiers.setX(fold, WUtil
                                        .getStrFromObj(current));
                                /* =================================================== */
                                Instances test = inst.testCV(numFolds, fold);

                                /* =============== Weka4WS code removed ============== */

                                for (int jj = 0; jj < test.numInstances(); jj++) {
                                    processClassifierPrediction(test
                                            .instance(jj), current, eval,
                                            predictions, predInstances,
                                            plotShape, plotSize);
                                    /* =============== Weka4WS code removed ============== */
                                }
                            }
                            /* ============== Weka4WS modifications ============== */
                            modelResult.setModels(classifiers);
                            /* =================================================== */
                            /* =============== Weka4WS code removed ============== */
                            break;

                        case 2: // Percent split
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Randomizing instances...");
                            rnd = evalOptions.getRnd();
                            /* =================================================== */
                            inst.randomize(new Random(rnd));
                            int trainSize = inst.numInstances() * percent / 100;
                            int testSize = inst.numInstances() - trainSize;
                            Instances train = new Instances(inst, 0, trainSize);
                            Instances test = new Instances(inst, trainSize,
                                    testSize);

                            /* ============== Weka4WS modifications ============== */
                            logger.info("Building model on training split...");
                            classifiers = new ArrayOfStrings(new String[1]);
                            /* =================================================== */

                            Classifier current = null;
                            try {
                                current = Classifier.makeCopy(template);
                            } catch (Exception ex) {
                                /* ============== Weka4WS modifications ============== */
                                logger.error("Problem copying classifier: "
                                        + ex.getMessage());
                                /* =================================================== */
                            }
                            current.buildClassifier(train);
                            eval = new Evaluation(train, costMatrix);

                            /* ============== Weka4WS modifications ============== */
                            classifiers.setX(0, WUtil.getStrFromObj(current));
                            modelResult.setModels(classifiers);
                            logger.info("Evaluating on test split...");
                            /* =================================================== */

                            /* =============== Weka4WS code removed ============== */

                            for (int jj = 0; jj < test.numInstances(); jj++) {
                                processClassifierPrediction(test.instance(jj),
                                        current, eval, predictions,
                                        predInstances, plotShape, plotSize);
                                /* =============== Weka4WS code removed  ============== */
                            }
                            /* =============== Weka4WS code removed  ============== */
                            break;

                        case 4: // Test on user split
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Evaluating on test data...");
                            /* =================================================== */

                            eval = new Evaluation(inst, costMatrix);

                            /* =============== Weka4WS code removed  ============== */

                            for (int jj = 0; jj < userTest.numInstances(); jj++) {
                                processClassifierPrediction(userTest
                                        .instance(jj), classifier, eval,
                                        predictions, predInstances, plotShape,
                                        plotSize);
                                /* =============== Weka4WS code removed ============== */
                            }
                            /* =============== Weka4WS code removed ============== */
                            break;

                        default:
                            throw new Exception("Test mode not implemented");
                        }

                        /* ============== Weka4WS modifications ============== */

                        visResult.setPlotShape(WUtil.getStrFromObj(plotShape));
                        visResult.setPlotSize(WUtil.getStrFromObj(plotSize));

                        // TODO predictions and predInstances may cause "Java Out of Memory Error" with large datasets
                        visResult.setPredictions(WUtil
                                .getStrFromObj(predictions));
                        visResult.setPredInstances(WUtil
                                .getStrFromObj(predInstances));

                        /* =================================================== */

                        if (outputSummary) {
                            /* ============== Weka4WS modifications ============== */
                            evalResult.setSummary(eval
                                    .toSummaryString(outputEntropy)
                                    + "\n");
                            /* =================================================== */
                        }

                        if (inst.attribute(classIndex).isNominal()) {

                            if (outputPerClass) {
                                /* ============== Weka4WS modifications ============== */
                                evalResult.setClassDetails(eval
                                        .toClassDetailsString()
                                        + "\n");
                                /* =================================================== */
                            }

                            if (outputConfusion) {
                                /* ============== Weka4WS modifications ============== */
                                evalResult.setMatrix(eval.toMatrixString()
                                        + "\n");
                                /* =================================================== */
                            }
                        }

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        logger.info("Finished " + cname);

                        result.setModel(modelResult);
                        result.setEvaluation(evalResult);
                        result.setVisualization(visResult);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                    } catch (Exception ex) {
                        /* ============== Weka4WS modifications ============== */
                        ex.printStackTrace();
                        wekaException.setThrown(true);
                        wekaException.setMessage(ex.getMessage());
                        /* =================================================== */
                    } finally {

                        /* =============== Weka4WS code removed ============== */

                        synchronized (this) {
                            /* =============== Weka4WS code removed ============== */
                            m_RunThread = null;
                            /* =============== Weka4WS code removed ============== */
                        }

                        /* ============== Weka4WS modifications ============== */
                        finalOperations(result, wekaException);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                    }
                }
            };
            /* =============== Weka4WS code removed ============== */

            m_RunThread.start();
        }
    }

    /**
     * Method taken from the ClassifierPanel class of Weka 3.4.11.
     */
    public static Instances setUpVisualizableInstances(Instances trainInstances) {
        FastVector hv = new FastVector();
        Attribute predictedClass;

        Attribute classAt = trainInstances.attribute(trainInstances
                .classIndex());
        if (classAt.isNominal()) {
            FastVector attVals = new FastVector();
            for (int i = 0; i < classAt.numValues(); i++) {
                attVals.addElement(classAt.value(i));
            }
            predictedClass = new Attribute("predicted" + classAt.name(),
                    attVals);
        } else {
            predictedClass = new Attribute("predicted" + classAt.name());
        }

        for (int i = 0; i < trainInstances.numAttributes(); i++) {
            if (i == trainInstances.classIndex()) {
                hv.addElement(predictedClass);
            }
            hv.addElement(trainInstances.attribute(i).copy());
        }
        return new Instances(trainInstances.relationName() + "_predicted", hv,
                trainInstances.numInstances());
    }

    /**
     * Method taken from the ClassifierPanel class of Weka 3.4.11
     */
    public static void processClassifierPrediction(Instance toPredict,
            Classifier classifier, Evaluation eval, FastVector predictions,
            Instances plotInstances, FastVector plotShape, FastVector plotSize) {
        try {
            double pred;
            // classifier is a distribution classifier and class is nominal
            if (predictions != null) {
                Instance classMissing = (Instance) toPredict.copy();
                classMissing.setDataset(toPredict.dataset());
                classMissing.setClassMissing();
                Classifier dc = classifier;
                double[] dist = dc.distributionForInstance(classMissing);
                pred = eval.evaluateModelOnce(dist, toPredict);
                predictions.addElement(new NominalPrediction(toPredict
                        .classValue(), dist, toPredict.weight()));
            } else {
                pred = eval.evaluateModelOnce(classifier, toPredict);
            }

            double[] values = new double[plotInstances.numAttributes()];
            for (int i = 0; i < plotInstances.numAttributes(); i++) {
                if (i < toPredict.classIndex()) {
                    values[i] = toPredict.value(i);
                } else if (i == toPredict.classIndex()) {
                    values[i] = pred;
                    values[i + 1] = toPredict.value(i);
                    /*
                     * // if the class value of the instances to predict is
                     * missing then // set it to the predicted value if
                     * (toPredict.isMissing(i)) { values[i+1] = pred; }
                     */
                    i++;
                } else {
                    values[i] = toPredict.value(i - 1);
                }
            }

            plotInstances.add(new Instance(1.0, values));
            if (toPredict.classAttribute().isNominal()) {
                if (toPredict.isMissing(toPredict.classIndex())
                        || Instance.isMissingValue(pred)) {
                    plotShape.addElement(new Integer(Plot2D.MISSING_SHAPE));
                } else if (pred != toPredict.classValue()) {
                    // set to default error point shape
                    plotShape.addElement(new Integer(Plot2D.ERROR_SHAPE));
                } else {
                    // otherwise set to constant (automatically assigned) point
                    // shape
                    plotShape.addElement(new Integer(
                            Plot2D.CONST_AUTOMATIC_SHAPE));
                }
                plotSize.addElement(new Integer(Plot2D.DEFAULT_SHAPE_SIZE));
            } else {
                // store the error (to be converted to a point size later)
                Double errd = null;
                if (!toPredict.isMissing(toPredict.classIndex())
                        && !Instance.isMissingValue(pred)) {
                    errd = new Double(pred - toPredict.classValue());
                    plotShape.addElement(new Integer(
                            Plot2D.CONST_AUTOMATIC_SHAPE));
                } else {
                    // missing shape if actual class not present or prediction
                    // is missing
                    plotShape.addElement(new Integer(Plot2D.MISSING_SHAPE));
                }
                plotSize.addElement(errd);
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

}
