/*
 *  ClusteringTask.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import java.io.File;
import java.io.FileReader;
import java.util.Random;

import weka.clusterers.ClusterEvaluation;
import weka.clusterers.Clusterer;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Remove;
import weka.gui.visualize.Plot2D;
import weka.gui.visualize.PlotData2D;
import weka4ws.stubs.service.ArrayOfStrings;
import weka4ws.stubs.service.EvalResult;
import weka4ws.stubs.service.ModelResult;
import weka4ws.stubs.service.TestOptions;
import weka4ws.stubs.service.VisResult;

import com.thoughtworks.xstream.XStream;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class ClusteringTask extends DataMiningTestTask {

    protected int classIndex;

    protected int[] ignoredAtts;

    public ClusteringTask(ClusteringParameters parameters,
            Weka4WSResource resultResource) throws Exception {

        super(parameters, "clustering", resultResource);

    }

    /**
     * This is the same method of the class
     * <code>weka.gui.explorer.ClustererPanel</code> taken from the source
     * code of Weka version 3.4.11
     */
    protected void startClusterer() {

        if (m_RunThread == null) {

            /* =============== Weka4WS code removed ============== */

            m_RunThread = new Thread() {
                public void run() {

                    /* =============== Weka4WS code removed ============== */

                    /* ============== Weka4WS modifications ============== */

                    try {
                        ClusteringParameters clustParameters = (ClusteringParameters) parameters;
                        TestOptions testOptions = clustParameters
                                .getTestOptions();

                        File datasetFile = new File(tempDataset.getFilePath());
                        FileReader datasetReader = new FileReader(datasetFile);
                        Instances inst = new Instances(datasetReader);

                        ModelResult modelResult = new ModelResult();
                        EvalResult evalResult = new EvalResult();
                        VisResult visResult = new VisResult();

                        classIndex = clustParameters.getClassIndex();
                        /* =================================================== */

                        inst.setClassIndex(classIndex);
                        Instances userTest = null;
                        PlotData2D predData = null;

                        /* ============== Weka4WS modifications ============== */

                        int percent = testOptions.getPercent();

                        if (testMode == TEST_SET) {
                            File testsetFile = new File(tempTestset
                                    .getFilePath());
                            FileReader testsetReader = new FileReader(
                                    testsetFile);
                            userTest = new Instances(testsetReader);
                        }

                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        ignoredAtts = clustParameters.getSelectedAttributes()
                                .getX();
                        if ((ignoredAtts != null) && (ignoredAtts.length == 0)) {
                            ignoredAtts = null;
                        }

                        String[] optionsArray = Utils.splitOptions(algorithm
                                .getParameters());

                        Clusterer clusterer = Clusterer.forName(algorithm
                                .getName(), optionsArray);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        String cname = clusterer.getClass().getName();

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        logger.info("Started " + cname);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        Instances trainInst = new Instances(inst);

                        /* =============== Weka4WS code removed ============== */

                        if (ignoredAtts != null) {
                            trainInst = removeIgnoreCols(trainInst);
                        }

                        /* =============== Weka4WS code removed ============== */

                        // Build the model and output it.
                        /* ============== Weka4WS modifications ============== */
                        logger.info("Building model on training data...");
                        /* =================================================== */

                        // remove the class attribute (if set) and build the
                        // clusterer
                        clusterer.buildClusterer(removeClass(trainInst));

                        /* ============== Weka4WS modifications ============== */
                        modelResult.setModel(WUtil.getStrFromObj(clusterer));
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        ArrayOfStrings clusterers;
                        /* =================================================== */

                        ClusterEvaluation eval = new ClusterEvaluation();
                        eval.setClusterer(clusterer);
                        switch (testMode) {
                        case 3:
                        case 5: // Test on training
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Evaluating on training data...");
                            /* =================================================== */

                            eval.evaluateClusterer(trainInst);
                            predData = setUpVisualizableInstances(inst, eval);

                            /* =============== Weka4WS code removed ============== */

                            break;

                        case 2: // Percent split
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Randomizing instances...");
                            /* =================================================== */

                            inst.randomize(new Random(1));
                            trainInst.randomize(new Random(1));
                            int trainSize = trainInst.numInstances() * percent
                                    / 100;
                            int testSize = trainInst.numInstances() - trainSize;
                            Instances train = new Instances(trainInst, 0,
                                    trainSize);
                            Instances test = new Instances(trainInst,
                                    trainSize, testSize);
                            Instances testVis = new Instances(inst, trainSize,
                                    testSize);

                            /* ============== Weka4WS modifications ============== */
                            logger.info("Building model on training split...");
                            /* =================================================== */

                            clusterer.buildClusterer(train);

                            /* ============== Weka4WS modifications ============== */

                            clusterers = new ArrayOfStrings(new String[1]);

                            clusterers.setX(0, WUtil.getStrFromObj(clusterer));

                            modelResult.setModels(clusterers);

                            logger.info("Evaluating on test split...");
                            /* =================================================== */

                            eval.evaluateClusterer(test);
                            predData = setUpVisualizableInstances(testVis, eval);

                            /* =============== Weka4WS code removed ============== */

                            break;

                        case 4: // Test on user split
                            /* ============== Weka4WS modifications ============== */
                            logger.info("Evaluating on test data...");
                            /* =================================================== */

                            Instances userTestT = new Instances(userTest);
                            if (ignoredAtts != null) {
                                userTestT = removeIgnoreCols(userTestT);
                            }
                            eval.evaluateClusterer(userTestT);
                            predData = setUpVisualizableInstances(userTest,
                                    eval);

                            /* =============== Weka4WS code removed ============== */

                            break;

                        default:
                            throw new Exception("Test mode not implemented");
                        }

                        /* =============== Weka4WS code removed ============== */

                        /* ============== Weka4WS modifications ============== */
                        evalResult.setSummary(eval.clusterResultsToString());

                        // TODO predData may cause "Java Out of Memory Error" with large datasets
                        XStream xstream = new XStream();
                        xstream.alias("plotdata2d", PlotData2D.class);
                        String xstreamedPredData = xstream.toXML(predData);
                        visResult.setPredInstances(xstreamedPredData);

                        logger.info("Finished " + cname);

                        result.setModel(modelResult);
                        result.setEvaluation(evalResult);
                        result.setVisualization(visResult);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                    } catch (Exception ex) {
                        /* ============== Weka4WS modifications ============== */
                        ex.printStackTrace();
                        wekaException.setThrown(true);
                        wekaException.setMessage(ex.getMessage());
                        /* =================================================== */
                    } finally {

                        /* =============== Weka4WS code removed ============== */

                        m_RunThread = null;

                        /* ============== Weka4WS modifications ============== */
                        finalOperations(result, wekaException);
                        /* =================================================== */

                        /* =============== Weka4WS code removed ============== */

                    }
                }
            };
            /* =============== Weka4WS code removed ============== */

            m_RunThread.start();
        }
    }

    /**
     * Method taken from the ClustererPanel class of Weka 3.4.11
     */
    public static PlotData2D setUpVisualizableInstances(
            Instances testInstances, ClusterEvaluation eval) throws Exception {

        int numClusters = eval.getNumClusters();
        double[] clusterAssignments = eval.getClusterAssignments();

        FastVector hv = new FastVector();
        Instances newInsts;

        Attribute predictedCluster;
        FastVector clustVals = new FastVector();

        for (int i = 0; i < numClusters; i++) {
            clustVals.addElement("cluster" + i);
        }
        predictedCluster = new Attribute("Cluster", clustVals);
        for (int i = 0; i < testInstances.numAttributes(); i++) {
            hv.addElement(testInstances.attribute(i).copy());
        }
        hv.addElement(predictedCluster);

        newInsts = new Instances(testInstances.relationName() + "_clustered",
                hv, testInstances.numInstances());

        double[] values;
        int j;
        int[] pointShapes = null;
        int[] classAssignments = null;
        if (testInstances.classIndex() >= 0) {
            classAssignments = eval.getClassesToClusters();
            pointShapes = new int[testInstances.numInstances()];
            for (int i = 0; i < testInstances.numInstances(); i++) {
                pointShapes[i] = Plot2D.CONST_AUTOMATIC_SHAPE;
            }
        }

        for (int i = 0; i < testInstances.numInstances(); i++) {
            values = new double[newInsts.numAttributes()];
            for (j = 0; j < testInstances.numAttributes(); j++) {
                values[j] = testInstances.instance(i).value(j);
            }
            values[j] = clusterAssignments[i];
            newInsts.add(new Instance(1.0, values));
            if (pointShapes != null) {
                if ((int) testInstances.instance(i).classValue() != classAssignments[(int) clusterAssignments[i]]) {
                    pointShapes[i] = Plot2D.ERROR_SHAPE;
                }
            }
        }
        PlotData2D plotData = new PlotData2D(newInsts);
        if (pointShapes != null) {
            plotData.setShapeType(pointShapes);
        }
        plotData.addInstanceNumberAttribute();
        return plotData;
    }

    /**
     * Method taken from the ClustererPanel class of Weka 3.4.11
     * 
     * @see weka.gui.explorer.ClustererPanel.removeClass
     */
    private Instances removeClass(Instances inst) {
        Remove af = new Remove();
        Instances retI = null;

        try {
            if (inst.classIndex() < 0) {
                retI = inst;
            } else {
                af.setAttributeIndices("" + (inst.classIndex() + 1));
                af.setInvertSelection(false);
                af.setInputFormat(inst);
                retI = Filter.useFilter(inst, af);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return retI;
    }

    /**
     * Method taken from the ClustererPanel class of Weka 3.4.11
     * 
     * @see weka.gui.explorer.ClustererPanel.removeClass
     */
    private Instances removeIgnoreCols(Instances inst) {

        // If the user is doing classes to clusters evaluation and
        // they have opted to ignore the class, then unselect the class in
        // the ignore list
        /* ============== Weka4WS modifications ============== */
        int[] selected;
        if (testMode == CLASSES_TO_CLUSTERS) {
            selected = remove(ignoredAtts, classIndex);
        } else {
            selected = ignoredAtts;
        }
        /* =================================================== */

        Remove af = new Remove();
        Instances retI = null;

        try {
            af.setAttributeIndicesArray(selected);
            af.setInvertSelection(false);
            af.setInputFormat(inst);
            retI = Filter.useFilter(inst, af);
        } catch (Exception e) {
            e.printStackTrace();
        }

        return retI;
    }

    private int numOccurrences(int[] array, int element) {
        int count = 0;
        for (int i = 0; i < array.length; i++) {
            if (array[i] == element) {
                count++;
            }
        }
        return count;
    }

    private int[] remove(int[] array, int element) {
        int num = numOccurrences(array, element);
        if (num >= 0) {
            int[] newArray = new int[array.length - num];
            int j = 0;
            for (int i = 0; i < array.length; i++) {
                if (array[i] != element) {
                    newArray[j++] = array[i];
                }
            }
        }
        return array;
    }

}
