/*
 *  DataMiningTestTask.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import weka4ws.stubs.service.Dataset;
import weka4ws.stubs.service.Result;
import weka4ws.stubs.service.Weka4WSException;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class DataMiningTestTask extends DataMiningTask {

    protected Dataset testset;

    protected Dataset tempTestset = new Dataset();

    protected int testMode;

    public DataMiningTestTask(ServiceTestParameters parameters, String type,
            Weka4WSResource resultResource) throws Exception {
        super(parameters, type, resultResource);
        testMode = parameters.getTestOptions().getTestMode();
        if (testMode == TEST_SET) {
            testset = parameters.getTestset();
            testset.setDirPath(dataset.getDirPath());
            testset.setFilePath(dataset.getDirPath() + testset.getFileName());
            dataset.setFilePath(dataset.getDirPath() + dataset.getFileName());
            logger.info("... with supplied test set '" + testset.getFileName()
                    + "'");
        }
    }

    protected void finalOperations(Result result, Weka4WSException wekaException) {
        if (testMode == TEST_SET) {
            String tempTestsetFilePath = tempTestset.getFilePath();
            removeFromTempDir(tempTestsetFilePath);
        }
        super.finalOperations(result, wekaException);
    }

    public boolean isTestsetAvailable() throws Exception {
        return isAvailable(testset, tempTestset);
    }

    /**
     * Stops the running task corresponding to the passed ID.
     */
    protected String stopTask() {

        if (m_RunThread != null) {
            if (testMode == TEST_SET) {
                removeFromTempDir(tempTestset.getFilePath());
            }
        }
        return super.stopTask();
    }
}
