/*
 *  HostCheckConnectionThread.java
 *  Copyright (C) 2005 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Writer;
import java.util.Date;

import org.apache.axis.encoding.Base64;

import weka.core.Instances;

/**
 * A collection of static methods and fields useful for Weka4WS Explorer and
 * Knowledge Flow component.
 * 
 * @author Oreste Verta
 * @author Marco Lackovic
 * @version $Revision: 1.0 $
 */
public class WUtil {

    public static String VERSION = "2.1";

    public static boolean logging = false;

    public static void copyFile(File in, File out) throws Exception {
        FileInputStream fis = new FileInputStream(in);
        FileOutputStream fos = new FileOutputStream(out);
        byte[] buf = new byte[1024];
        int i = 0;
        while ((i = fis.read(buf)) != -1) {
            fos.write(buf, 0, i);
        }
        fis.close();
        fos.close();
    }

    /**
     * Saves the instances in a file, located in the user's hidden subfolder
     * named ".Weka4WS", having the name of the original dataset plus (for
     * uniquity) the number of milliseconds since January 1, 1970.
     */
    public static String saveWorkingInstancesToFile(String originalFilePath,
            Instances instances) {

        String fileName = originalFilePath.substring(originalFilePath
                .lastIndexOf(File.separator) + 1, originalFilePath
                .lastIndexOf("."));
        fileName += new Date().getTime() + Instances.FILE_EXTENSION;
        String saveDir = System.getProperty("user.home") + File.separator
                + ".Weka4WS";
        if (!(new File(saveDir)).exists()) {
            (new File(saveDir)).mkdir();
        }
        String newDatasetFilePath = saveDir + File.separator + fileName;
        File newDatasetFile = new File(newDatasetFilePath);
        saveInstancesToFile(newDatasetFile, instances);
        return newDatasetFilePath;
    }

    /**
     * Saves the current instances to the supplied file.
     * 
     * @param f
     *            a value of type 'File'
     * @param inst
     *            the instances to save
     * @param saveHeader
     *            true to save in arff format, false to save in csv
     */
    private static void saveInstancesToFile(final File f, final Instances inst) {

        try {
            Writer w = new BufferedWriter(new FileWriter(f));

            Instances h = new Instances(inst, 0);
            w.write(h.toString());
            w.write("\n");

            for (int i = 0; i < inst.numInstances(); i++) {
                w.write(inst.instance(i).toString());
                w.write("\n");
            }
            w.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    public static Object getObjFromStr(String modelString) {

        if ((modelString == null) || (modelString == "")
                || (modelString.length() == 0)) {
            return null;
        }

        Object model = null;
        ByteArrayInputStream byteStream = new ByteArrayInputStream(Base64
                .decode(modelString));
        ObjectInputStream inputStream = null;
        try {
            inputStream = new ObjectInputStream(byteStream);
            model = inputStream.readObject();
            inputStream.close();
        } catch (IOException ex) {
            ex.printStackTrace();
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
        }

        return model;
    }

    public static String getStrFromObj(Object model) {
        if (model == null) {
            return "";
        }
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        ObjectOutputStream outputStream = null;
        String encodedString = null;

        try {
            outputStream = new ObjectOutputStream(byteStream);
            outputStream.writeObject(model);
            outputStream.close();
            encodedString = Base64.encode(byteStream.toByteArray());
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        return encodedString;
    }

    public static void waitForAnyKeyToContinue() throws IOException {
        BufferedReader stdin = new BufferedReader(new InputStreamReader(
                System.in));
        stdin.read();
    }

    public static void setLogging(boolean logging) {
        WUtil.logging = logging;
    }

}
