/*
 *  Weka4WSService.java
 *  Copyright (C) 2005 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import java.util.Hashtable;

import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceContextException;
import org.globus.wsrf.ResourceException;

import weka4ws.stubs.service.Response;
import weka4ws.stubs.service.Result;
import weka4ws.stubs.service.StopParameters;
import weka4ws.stubs.service.StopResponse;
import weka4ws.stubs.service.VersionParameters;
import weka4ws.stubs.service.Weka4WSException;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class Weka4WSService implements DataMiningConstants {

    protected static Hashtable runningTasks = new Hashtable();

    /**
     * Submits an association rules task.
     */
    public Response associationRules(AssociationRulesParameters parameters) {

        WUtil.setLogging(parameters.isLogging());

        Weka4WSException wekaException = new Weka4WSException();
        wekaException.setThrown(false);

        Response response = new Response();
        response.setException(wekaException);

        AssociationRulesTask task;

        try {
            task = new AssociationRulesTask(parameters, getResultResource());

            if (task.isDatasetAvailable()) {
                task.startAssociator();
                runningTasks.put(new Long(parameters.getTaskID()), task);
                response.setDatasetFound(true);
                return response;
            }
            response.setDatasetFound(false);
            response.setDirPath(task.getSavePath());
            return response;
        } catch (Exception e) {
            wekaException.setThrown(true);
            wekaException.setMessage(e.getMessage());
            logger.error(e.getMessage());
            return response;
        }
    }

    /**
     * Submits a classification task.
     */
    public Response classification(ClassificationParameters parameters) {

        WUtil.setLogging(parameters.isLogging());

        // SecurityManager.getManager().setServiceOwnerFromContext();

        Weka4WSException wekaException = new Weka4WSException();
        wekaException.setThrown(false);

        Response response = new Response();
        response.setException(wekaException);
        ClassificationTask task;
        try {
            task = new ClassificationTask(parameters, getResultResource());

            if (parameters.getTestOptions().getTestMode() == TEST_SET) {
                if (task.isDatasetAvailable()) {
                    response.setDatasetFound(true);
                    if (task.isTestsetAvailable()) {
                        response.setTestsetFound(true);
                        task.startClassifier();
                        runningTasks
                                .put(new Long(parameters.getTaskID()), task);
                    } else {
                        response.setTestsetFound(false);
                        response.setDirPath(task.getSavePath());
                    }
                    return response;
                }
                response.setDatasetFound(false);
                if (task.isTestsetAvailable()) {
                    response.setTestsetFound(true);
                } else {
                    response.setTestsetFound(false);
                }
                response.setDirPath(task.getSavePath());
                return response;
            } else if (task.isDatasetAvailable()) {
                task.startClassifier();
                runningTasks.put(new Long(parameters.getTaskID()), task);
                response.setDatasetFound(true);
                return response;
            }
            response.setDatasetFound(false);
            response.setDirPath(task.getSavePath());
            return response;
        } catch (Exception e) {
            wekaException.setThrown(true);
            wekaException.setMessage(e.getMessage());
            logger.error(e.getMessage());
            return response;
        }
    }

    /**
     * Submits a clustering task.
     */
    public Response clustering(ClusteringParameters parameters) {

        WUtil.setLogging(parameters.isLogging());

        Weka4WSException wekaException = new Weka4WSException();
        wekaException.setThrown(false);

        Response response = new Response();
        response.setException(wekaException);
        ClusteringTask task;
        try {
            task = new ClusteringTask(parameters, getResultResource());

            if (parameters.getTestOptions().getTestMode() == TEST_SET) {
                if (task.isDatasetAvailable()) {
                    response.setDatasetFound(true);
                    if (task.isTestsetAvailable()) {
                        response.setTestsetFound(true);
                        task.startClusterer();
                        runningTasks
                                .put(new Long(parameters.getTaskID()), task);
                    } else {
                        response.setTestsetFound(false);
                        response.setDirPath(task.getSavePath());
                    }
                    return response;
                }
                response.setDatasetFound(false);
                if (task.isTestsetAvailable()) {
                    response.setTestsetFound(true);
                } else {
                    response.setTestsetFound(false);
                }
                response.setDirPath(task.getSavePath());
                return response;
            } else if (task.isDatasetAvailable()) {
                task.startClusterer();
                runningTasks.put(new Long(parameters.getTaskID()), task);
                response.setDatasetFound(true);
                return response;
            }
        } catch (Exception e) {
            wekaException.setThrown(true);
            wekaException.setMessage(e.getMessage());
            logger.error(e.getMessage());
            return response;
        }
        response.setDatasetFound(false);
        response.setDirPath(task.getSavePath());
        return response;
    }

    /**
     * Returns the Weka4WS resource containing the data mining task results
     */
    /* WARNING: this method must stay inside this class */
    protected Weka4WSResource getResultResource()
            throws ResourceContextException, ResourceException {
        Object resource = ResourceContext.getResourceContext().getResource();
        Weka4WSResource wekaResource = (Weka4WSResource) resource;
        return wekaResource;
    }

    /**
     * Stops the given running task (if any).
     */
    public StopResponse stopTask(StopParameters parameters) {
        logger.info("Interruption requested by the client");

        Long key = new Long(parameters.getTaskID());
        DataMiningTask task = (DataMiningTask) runningTasks.get(key);
        StopResponse response = new StopResponse();
        if (task != null) {
            response.setResponse(task.stopTask());
            task = null;
        } else {
            logger.info("Task to stop not found");
        }
        runningTasks.remove(key);
        return response;
    }

    public String getVersion(VersionParameters parameters) {
        return WUtil.VERSION;
    }

    public void notifCheck(VersionParameters parameters)
            throws ResourceContextException, ResourceException {
        Object resource = ResourceContext.getResourceContext().getResource();
        Weka4WSResource wekaResource = (Weka4WSResource) resource;
        wekaResource.setResult(new Result());
    }

    protected static synchronized void removeTask(long key) {
        runningTasks.remove(new Long(key));
    }
}
